package fe

import (
	"log"

	"git.autistici.org/ale/radioai"
	"github.com/jmcvetta/randutil"
)

// A load balancing policy selects a single node from the pool of
// currently active ones.
type LoadBalancingPolicy interface {
	GetNode([]*radioai.NodeStatus) *radioai.NodeStatus
}

// Simple load balancing policy that always returns the nodes with the
// least amount of listeners.
type leastListenersPolicy struct{}

func (llp leastListenersPolicy) GetNode(nodes []*radioai.NodeStatus) *radioai.NodeStatus {
	minIdx := 0
	min := 1000000
	for i, n := range nodes {
		if listeners := n.NumListeners(); listeners < min {
			minIdx = i
			min = listeners
		}
	}
	return nodes[minIdx]
}

// Simple load balancing policy that selects a node randomly,
// associating a weight to each node inversely proportional to the
// number of active listeners.
type weightedListenersPolicy struct{}

func (wlp weightedListenersPolicy) GetNode(nodes []*radioai.NodeStatus) *radioai.NodeStatus {
	choices := make([]randutil.Choice, 0, len(nodes))
	weightBase := 1000000
	for _, n := range nodes {
		w := weightBase / (n.NumListeners() + 1)
		choices = append(choices, randutil.Choice{w, n})
	}
	result, err := randutil.WeightedChoice(choices)
	if err != nil {
		return nil
	}
	return result.Item.(*radioai.NodeStatus)
}

func getNamedLoadBalancingPolicy(lbpolicy string) LoadBalancingPolicy {
	switch lbpolicy {
	case "leastloaded", "ll":
		return &leastListenersPolicy{}
	case "weighted", "wl":
		return &weightedListenersPolicy{}
	}
	log.Fatalf("Unknown load-balancing policy '%s'", lbpolicy)
	return nil
}
