package node

import (
	"encoding/xml"
	"errors"
	"io"
	"log"
	"net/http"
	"os"
	"os/exec"
	"strconv"
	"time"

	"git.autistici.org/ale/radioai"
)

var (
	statusPageUrl = "http://localhost:8000/status-radioai.xsl"
)

// Icecast returns empty fields in our status handler, which we'll
// need to turn into integers (the xml unmarshaler will return an
// error in this specific case), so we use a separate type for
// decoding the status page output. This would be much simpler if I
// knew how to get the XSLT to put a default value in the output
// instead of an empty field...
type icecastMountStatusUnparsed struct {
	Name         string `xml:"name,attr"`
	Listeners    string `xml:"listeners"`
	BitRate      string `xml:"bitrate"`
	Quality      string `xml:"quality"`
	VideoQuality string `xml:"video-quality"`
	FrameSize    string `xml:"frame-size"`
	FrameRate    string `xml:"frame-rate"`
}

type icecastStatusUnparsed struct {
	XMLName xml.Name                     `xml:"status"`
	Mounts  []icecastMountStatusUnparsed `xml:"mount"`
}

type IcecastStatus struct {
	Mounts []radioai.IcecastMountStatus
	Up     bool
}

type IcecastController struct {
	PublicIp   string
	ConfigFile string
	InitScript string
	config     *icecastConfig
	status     *IcecastStatus
	stop       chan bool
}

func NewIcecastController(publicIp string, stop chan bool) *IcecastController {
	return &IcecastController{
		PublicIp:   publicIp,
		ConfigFile: "/etc/icecast2/icecast.xml",
		InitScript: "/etc/init.d/icecast2",
		config:     newIcecastConfig(publicIp),
		status:     &IcecastStatus{},
		stop:       make(chan bool, 1),
	}
}

func (ic *IcecastController) reload() error {
	err := exec.Command(ic.InitScript, "reload").Run()
	if err != nil {
		err = exec.Command(ic.InitScript, "start").Run()
	}
	return err
}

// Update reloads the Icecast daemon with a new configuration.
func (ic *IcecastController) Update(conf *ClusterConfig, isMaster bool, masterAddr string) error {
	if !isMaster && masterAddr == "" {
		return errors.New("unknown system state")
	}

	tmpf := ic.ConfigFile + ".tmp"
	defer os.Remove(tmpf)

	ic.config.Update(conf, isMaster, masterAddr)
	if err := ic.config.EncodeToFile(tmpf); err != nil {
		return err
	}

	if err := os.Rename(tmpf, ic.ConfigFile); err != nil {
		return err
	}

	return ic.reload()
}

func (ic *IcecastController) GetStatus() *IcecastStatus {
	return ic.status
}

func (ic *IcecastController) statusUpdater() {
	t := time.NewTicker(3 * time.Second)
	downStatus := &IcecastStatus{}
	for {
		select {
		case <-t.C:
			if status, err := ic.fetchStatus(); err == nil {
				ic.status = status
			} else {
				log.Printf("bad status from iceast: %v", err)
				ic.status = downStatus
			}
		case <-ic.stop:
			return
		}
	}
}

func (ic *IcecastController) fetchStatus() (*IcecastStatus, error) {
	resp, err := http.Get(statusPageUrl)
	if err != nil {
		return nil, err
	}
	defer resp.Body.Close()
	return ic.parseStatusPage(resp.Body)
}

func (ic *IcecastController) parseStatusPage(input io.Reader) (*IcecastStatus, error) {
	var ustatus icecastStatusUnparsed
	if err := xml.NewDecoder(input).Decode(&ustatus); err != nil {
		return nil, err
	}

	// Quick converters from string that default to 0.
	toi := func(s string) int {
		if i, err := strconv.Atoi(s); err == nil {
			return i
		}
		return 0
	}
	tof := func(s string) float64 {
		if f, err := strconv.ParseFloat(s, 64); err == nil {
			return f
		}
		return 0
	}

	status := IcecastStatus{
		Up:     true,
		Mounts: make([]radioai.IcecastMountStatus, 0, len(ustatus.Mounts)),
	}
	for _, um := range ustatus.Mounts {
		m := radioai.IcecastMountStatus{
			Name:         um.Name,
			Listeners:    toi(um.Listeners),
			BitRate:      toi(um.BitRate),
			Quality:      tof(um.Quality),
			VideoQuality: tof(um.VideoQuality),
			FrameSize:    um.FrameSize,
			FrameRate:    tof(um.FrameRate),
		}
		status.Mounts = append(status.Mounts, m)
	}

	return &status, nil
}

func (ic *IcecastController) Run() {
	ic.statusUpdater()
}
