package node

import (
	"bytes"
	"encoding/json"
	"log"
	"strings"
	"sync"
	"time"

	"git.autistici.org/ale/radioai"
	"git.autistici.org/ale/radioai/masterelection"
	"git.autistici.org/ale/radioai/third_party/github.com/coreos/go-etcd/etcd"
)

func trigger(c chan bool) {
	select {
	case c <- true:
	default:
	}
}

// Remove mountPrefix from the beginning of the path, but keep the
// leading slash.
func keyToMount(key string) string {
	return key[len(radioai.MountPrefix)-1:]
}

// In-memory representation of the overall configuration (basically
// just a list of the known mounts).
type ClusterConfig struct {
	mounts map[string]*radioai.Mount
	lock   sync.Mutex
}

func NewClusterConfig() *ClusterConfig {
	return &ClusterConfig{
		mounts: make(map[string]*radioai.Mount),
	}
}

// TODO: remove?
func (c *ClusterConfig) GetMount(name string) *radioai.Mount {
	c.lock.Lock()
	defer c.lock.Unlock()
	return c.mounts[name]
}

// TODO: remove?
func (c *ClusterConfig) ListMounts() []*radioai.Mount {
	c.lock.Lock()
	defer c.lock.Unlock()
	result := make([]*radioai.Mount, 0, len(c.mounts))
	for _, m := range c.mounts {
		result = append(result, m)
	}
	return result
}

// Update a mount (in-memory only).
func (c *ClusterConfig) setMount(m *radioai.Mount) {
	c.lock.Lock()
	defer c.lock.Unlock()
	c.mounts[m.Name] = m
}

// Delete a mount (in-memory only).
func (c *ClusterConfig) delMount(name string) {
	c.lock.Lock()
	defer c.lock.Unlock()
	delete(c.mounts, name)
}

// Keeps the in-memory service configuration in sync with the
// distributed database. An update channel is triggered whenever the
// data changes.
type ConfigSyncer struct {
	client *etcd.Client
	config *ClusterConfig
	rch    chan *etcd.Response
	upch   chan bool
	stop   chan bool
	index  uint64
}

func NewConfigSyncer(client *etcd.Client, config *ClusterConfig, upch chan bool, stop chan bool) *ConfigSyncer {
	return &ConfigSyncer{
		client: client,
		config: config,
		rch:    make(chan *etcd.Response),
		upch:   upch,
		stop:   stop,
	}
}

func (w *ConfigSyncer) syncer() {
	log.Printf("starting syncer")
	for {
		select {
		case response := <-w.rch:

			if response.Action == "delete" {
				mountName := keyToMount(response.Key)
				log.Printf("deleted mount %s", mountName)
				w.config.delMount(mountName)
			} else if response.Action == "set" || response.Action == "create" || response.Action == "update" {
				w.updateConfigWithResponse(response.Key, response.Value)
			} else {
				continue
			}

			// Update the 'last seen' index, so that if
			// the Watcher dies, it knows where to start
			// from and we do not have to download the
			// full configuration again.
			w.index = response.ModifiedIndex

			// Trigger an update.
			trigger(w.upch)

		case <-w.stop:
			return
		}
	}
}

func (w *ConfigSyncer) updateConfigWithResponse(key, value string) {
	mountName := keyToMount(key)
	log.Printf("updating mount %s: %s", mountName, value)
	var m radioai.Mount
	if err := json.NewDecoder(strings.NewReader(value)).Decode(&m); err != nil {
		log.Printf("corrupted data: %s: %s", value, err)
	} else {
		w.config.setMount(&m)
	}
}

// Run the ConfigSyncer in the background. It will wait for
// initialization to complete, so that when this function returns, the
// in-memory configuration has already been fully synchronized.
func (w *ConfigSyncer) Run() {
	// Run until the first successful Get().
	log.Printf("attempting to retrieve initial config...")
	for {
		response, err := w.client.Get(radioai.MountPrefix, false)
		if err == nil && response.Dir {
			// Directly update the configuration.
			for _, r := range response.Kvs {
				w.updateConfigWithResponse(r.Key, r.Value)
			}
			w.index = response.ModifiedIndex
			break
		}
		log.Printf("Get error: %s", err)

		// Wait 1 second, but watch the stop channel.
		select {
		case <-time.After(1 * time.Second):
		case <-w.stop:
			return
		}
	}

	// Update the icecast daemon now that we have a full config.
	log.Printf("triggering initial reload")
	trigger(w.upch)

	// Now start the watcher, and the syncer.
	go w.syncer()
	go func() {
		for {
			curIndex := w.index + 1
			log.Printf("starting watcher at index %d", curIndex)
			_, err := w.client.WatchAll(radioai.MountPrefix, curIndex, w.rch, w.stop)
			if err == etcd.ErrWatchStoppedByUser {
				return
			} else if err != nil {
				// Log the error, but keep watching.
				log.Printf("Watch(): %s", err)
			}
		}
	}()
}

// An active streaming node, managing the local icecast server.
type RadioNode struct {
	Config *ClusterConfig

	ip          string
	client      *etcd.Client
	me          *masterelection.MasterElection
	watcher     *ConfigSyncer
	icecast     *IcecastController
	livenessTtl uint64
	upch        chan bool
	stop        chan bool
}

func NewRadioNode(ip string, client *etcd.Client) *RadioNode {
	config := NewClusterConfig()

	// Network updates trigger icecast reconfiguration. This
	// channel is used as an 'event', no more than one entry will
	// be queued.
	upch := make(chan bool, 1)

	// MasterElection changes trigger an update.
	mech := make(chan int)
	go func() {
		for _ = range mech {
			trigger(upch)
		}
	}()

	// Global 'stop' channel.
	stopch := make(chan bool)

	return &RadioNode{
		Config: config,
		ip:     ip,
		client: client,
		me: masterelection.NewMasterElection(
			client,
			radioai.MasterElectionPath,
			ip,
			5,
			mech,
			stopch),
		watcher:     NewConfigSyncer(client, config, upch, stopch),
		icecast:     NewIcecastController(ip, stopch),
		livenessTtl: 2,
		upch:        upch,
		stop:        stopch,
	}
}

// The presence goroutine continuously updates our entry in the list
// of nodes.
func (rc *RadioNode) presence() {
	ticker := time.NewTicker(time.Duration(rc.livenessTtl/2) * time.Second)

	for {
		select {
		case <-ticker.C:
			// Build our NodeStatus.
			icecastSt := rc.icecast.GetStatus()
			nodeSt := radioai.NodeStatus{
				IP: rc.ip,
				IcecastUp: icecastSt.Up,
				Mounts: icecastSt.Mounts,
			}

			// Update our node entry in the database.
			var buf bytes.Buffer
			json.NewEncoder(&buf).Encode(&nodeSt)
			if _, err := rc.client.Set(radioai.NodePrefix+rc.ip, buf.String(), rc.livenessTtl); err != nil {
				log.Printf("presence: Set(): %s", err)
			}

		case <-rc.stop:
			return
		}
	}
}

// Run the node. This method does not return until someone calls Stop().
func (rc *RadioNode) Run() {
	// Bootstrap the config watcher. This ensures that we have a
	// full configuration (thanks to the Get() call) before we
	// start managing the icecast server.
	rc.watcher.Run()

	// Start the presence heartbeat.
	go rc.presence()

	// Start the masterelection runner.
	go rc.me.Run()

	// Wait an instant to give a chance to the other services to
	// initialize.
	time.Sleep(200 * time.Millisecond)

	// Start the Icecast status checker.
	go rc.icecast.Run()

	log.Printf("starting icecast updater")
	for {
		select {
		case <-rc.upch:
			log.Printf("reloading icecast config")
			if err := rc.icecast.Update(rc.Config, rc.me.IsMaster(), rc.me.GetMasterAddr()); err != nil {
				log.Printf("Update(): %s", err)
			}

			// Do not reload icecast more often than once
			// every two seconds.
			time.Sleep(2 * time.Second)

		case <-rc.stop:
			return
		}
	}
}

// Stop everything.
func (rc *RadioNode) Stop() {
	close(rc.stop)

	// We should use WaitGroups here. Instead, wait 2 seconds.
	time.Sleep(2 * time.Second)
}
